# info: contains the functions for graph_numba.py

from numba import jit, njit, vectorize
import time
import matplotlib.pyplot as plt
import numpy as np

from concurrent.futures import ThreadPoolExecutor

from numba import cuda

from . import funcDictionary as dic
#import funcDictionaryLevenshtein as dicLeven
from . import plotData
import textdistance as td

import psutil
from scipy import sparse
import pandas as pd
import math

def isNaN(string):
    return string!=string

def convert_to_ascii(strings1, strings2, max_len):
    """
    info: take two arrays of strings and convert both into 
        the special ascii format, where each character
        is saved as an ubyte (unsigned byte), which is 8 
        bits.
        Actually ASCII would only need 7 bits, so there would
        still be little space for improvement.
        It is assumed, that no string is longer than max_len, which
        is usually 30. If a string is shorter, then "0"-characters
        will be appended until the length is max_len.
        Thus, all strings have the same length, which is 
        necessary in order to define equidistant buffers later
        for the GPU.
        (I think also there is a way to improve that.)
    inputs: strings1: array of strings
        strings2: array of strings
        max_len: maximum length, that each string can have 
        (usually 30)
    outputs: ascii_total1, ascii_total2:
            two string arrays converted to the special ascii_type, where
            each element has the same length max_len
    """

    with_log = False

    #ascii_total = np.array([])
    ascii_total = list()
    #strings = strings.astype(np.float32)

    for i in range(len(strings1)):
        s = strings1[i]
        if len(s) > max_len:
            if with_log:
                print("\n info: an s is longer then max_len, we cut it")
            s = s[:max_len]
        for j in range(len(s), max_len):
            s = s + '0'
        ascii_code = [ord(c) for c in s]
        for j in range(len(ascii_code)):
            ascii_total.append(ascii_code[j])
        #ascii_total = ascii_total + ascii_code
    ascii_total = np.array(ascii_total)
    ascii_total1 = ascii_total.astype(np.ubyte)

    ascii_total = list()
    #strings = strings.astype(np.float32)

    for i in range(len(strings2)):
        s = strings2[i]
        if len(s) > max_len:
            if with_log: 
                print("\n info: an s is longer then max_len, we cut it")
            s = s[:max_len]
        for j in range(len(s), max_len):
            s = s + '0'
        ascii_code = [ord(c) for c in s]
        for j in range(len(ascii_code)):
            ascii_total.append(ascii_code[j])
        #ascii_total = ascii_total + ascii_code
    ascii_total = np.array(ascii_total)
    ascii_total2 = ascii_total.astype(np.ubyte)

    return ascii_total1, ascii_total2

@cuda.jit('void(int8[:], int8[:], int32, int8[:], int8)')
def levenshtein(strings1, strings2, num_strings, dist, max_ld):
    """
    Info: GPU kernel - computes the Levenshtein distance based adjacency 
        matrix for the string lists strings1 and strings2
    Args: strings1: list of strings, where each string has 30 letters 
            in ascii format
        strings2: list of strings, where each string has 30 letters 
            in ascii format
        num_strings: number of strings in each of the arrays strings1 and strings2
            (hence we assume, that strings1 and strings2 have the same length)
        dist: empty buffer for the adjacency matrix
        max_ld: threshold distance
    Returns: -
    """ 

    test = 0

    tx = cuda.threadIdx.x
    ty = cuda.blockIdx.x
    bw = cuda.blockDim.x
    gid = ty*bw + tx
    
    lx = int(gid%num_strings)
    ly = int(gid/num_strings)

    m = 30
    n = 30
    m_done = False
    n_done = False
    
    dprev = cuda.local.array(shape=(30), dtype=np.int8)
    dnew = cuda.local.array(shape=(30), dtype=np.int8)
    els = cuda.local.array(shape=(3), dtype=np.int8)

    dp0, dp1, dp2, dp3, dp4, dp5, dp6, dp7, dp8, dp9, dp10, dp11, dp12, dp13, dp14, dp15, dp16, dp17, dp18, \
            dp19, dp20, dp21, dp22, dp23, dp24, dp25, dp26, dp27, dp28, dp29 = 0, 1, 2, 3, 4, 5, 6, 7, 8, \
            9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29

    dn0, dn1, dn2, dn3, dn4, dn5, dn6, dn7, dn8, dn9, dn10, dn11, dn12, dn13, dn14, dn15, dn16, dn17, dn18, \
            dn19, dn20, dn21, dn22, dn23, dn24, dn25, dn26, dn27, dn28, dn29 = 0, 1, 2, 3, 4, 5, 6, 7, 8, \
            9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29
    
    dif = cuda.local.array(shape=1, dtype=np.int8)
    strings1_local = cuda.local.array(shape=30, dtype=np.int8)
    strings2_local = cuda.local.array(shape=30, dtype=np.int8)

    for i in range(30):
        strings1_local[i] = strings1[lx*30 + i]
        strings2_local[i] = strings2[ly*30 + i]
        if strings1[lx*30 + i] == 48:
            if m_done == False:
                m = i
                m_done = True
            else:
                pass
        if strings2[ly*30 + i] == 48:
            if n_done == False:
                n = i
                n_done = True
            else:
                pass
    
    for j in range(m+1):
        dprev[j] = j
    d_val = -1
    
    min_val = 0

    for i in range(1, n+1):#n+1):
        dnew[0] = i
        dn0 = i

        min_val = 30
        for j in range(1, m+1):
            if strings2_local[i-1] == strings1_local[j-1]:
                substitutionCost = 0
            else:
                substitutionCost = 1
            """
            el1 = dprev[j]# +1
            el2 = dnew[j-1]# +1
            el3 = dprev[j-1] + substitutionCost
            
            if el2 < el1:
                el1 = el2
            if el3 < el1 + 1:
                el1 = el3
            else: 
                el1 += 1
            dnew[j] = el1
            """

            els[0] = dprev[j]# +1
            els[1] = dnew[j-1]# +1
            els[2] = dprev[j-1] + substitutionCost

            if els[1] < els[0]:
                els[0] = els[1]
            if els[2] < els[0] + 1:
                els[0] = els[2]
            else:
                els[0] += 1
            dnew[j] = els[0]
            
            
            if dnew[j] < min_val:
                min_val = dnew[j]
            
            #d_val = el1
        
        dprev[0] = i 
        
        for j in range(1, m+1):
            dprev[j] = dnew[j]
        
        if min_val > max_ld:
            break
        #print("\n A gid: ", gid, "; i: ", i, "; j: ", j, "; in min_val")

    #dist[gid] = d_val

    dist[gid] = 1 if dnew[m]<max_ld else 0

def insert_pathogens(seq, pathogen_name):
    """
    Info: replace the first strings in "seq" by the pathogen cognate CDR3 
        of the pathogen with name "pathogen_name", where the corresponding 
        cognate CDR3 are read from the "virus_databank.csv" file
    Args: seq: normal array of strings of CDR3 sequences (normal format)
        pathogen_name: name of the pathogen
    Returns: seq: new sequence with the replaced CDR3
        idx: number of replaced sequences
    """
   
    data = pd.read_csv('virus_databank.csv', low_memory=False)
    CDR3_beta_total = list(data["CDR3.beta.aa"])
    CDR3_beta_names = list(data["Pathology"])
    
    idx = 0
    for i in range(len(CDR3_beta_total)):
        #print("\n i: ", i, "; CDR3_beta_names[i]: ", CDR3_beta_names[i], "; pathogen_name: ", pathogen_name)
        if CDR3_beta_names[i] == pathogen_name:
            #print("\n has a chance at i=", i)
            #print("\n i: ", i, "; CDR3_beta_names[i]: ", CDR3_beta_names[i], "; pathogen_name: ", pathogen_name)
            #-try:
            if True: 
                #print("\n CDR3_beta_total[i]: ", CDR3_beta_total[i], "; isnan: ", isNaN(CDR3_beta_total[i]))
                #if not math.isnan(CDR3_beta_total[i]):
                if not isNaN(CDR3_beta_total[i]):
                    seq[idx] = CDR3_beta_total[i]
                    idx += 1
            #-except: 
            #-    pass
    # info: increasing by one in order to return a length value, 
    #    which is usually 1 higher than the highest index 
    idx += 1
    return seq, idx

def contract_array(h_total, idx_max, side_len):
    """
    Info: remove the edges, which just connect two antigen cognate CDR3
    Args: h_total: sparse adjacency matrix
        idx_max: number of sequences which are replaced by antigen cognate 
            CDR3 sequence
        side_len: length of string arrays, whose distance matrix is calculated
            (i.e., side length of distance matrix)
    Returns: new sparse adjacency matrix with removed strings
    """
    # perh. TODO: change the name to something more fitting, e.g. "remove self_edges"
    
    i = 0
    while i < len(h_total):
        x = h_total[i]%side_len
        y = int(h_total[i]/side_len)
        if (x < idx_max and y < idx_max):
            h_total = np.delete(h_total, i)
            i-=1
            #-print("\n the was a self edge among the pathogens")
        i+=1
    return h_total



# info: make graphs for the presentation

import numpy as np
import matplotlib.pyplot as plt
import networkx as nx

def make_clusters():
    """
    info: 
    input: 
    output: 
    """
    fig1, ax1 = plt.subplots()
    fig2, ax2 = plt.subplots()
    g1 = nx.gnp_random_graph(20, 0.1)
    
    num_nodes = len(list(g1.nodes()))
    cluster_idxs = list([None]*num_nodes)
    print("\n clusters_idxs: ", cluster_idxs)
    a = list(nx.connected_components(g1))
    for i in range(len(a)):
        for j in range(len(a[i])):
            cluster_idxs[list(a[i])[j]] = i/len(a)
    
    print("\n c: ", list(a))
    print("\n cluster_idxs: ", cluster_idxs)
    cmap = [(0.0, el, 0.0, 1.0) for el in cluster_idxs]
    
    nx.draw(g1, ax=ax1, node_color=cmap)
    plt.show()

def make_cliques():
    """
    info:
    input: 
    output: 
    """
    fig1, ax1 = plt.subplots()
    
    g = nx.gnp_random_graph(20, 0.2)
    
    num_nodes = len(list(g.nodes()))
    cliques_idxs = list([None]*num_nodes)
    """
    a = list(nx.find_cliques(g))
    print("\n a: ", a)
    for i in [0]:#range(len(a)):
        for j in range(len(a[i])):
            cliques_idxs[list(a[i])[j]] = 1#i/len(a)
    """
    
    cliques_idxs = list(nx.max_clique(g))
    print("\n cliques_idxs: ", cliques_idxs)
    for i in range(len(cliques_idxs)):
        if cliques_idxs[i] == None:
            cliques_idxs[i] = 0
        else: 
            cliques_idxs[i] = 1
    print("\n cliques_idxs: ", cliques_idxs)
    cmap = [(0.0, el, 0.0, 1.0) for el in cliques_idxs]
    nx.draw(g, ax=ax1, node_color=cmap)
    plt.show()

if __name__ == "__main__":
    #make_clusters()
    make_cliques()



import numpy as np
import networkx as nx
import collections
#import jaccardUtil
import csv

def make_nets(ls, seq, lMax=1000): 
    """
    info: make a list "nets", where each element "net" is 
        a list again, containing all sequences with a certain 
        length
    input:
        ls:  list of length values for each sequence in seq
        seq: array of nucleic acid sequences 
        lmax: maximum allowed number of nodes in each graph
    output: 
        nets: list of nets, where each element "net" is 
        a list again, containing all sequences with a certain 
        length  
    """
    lsHist = np.histogram(ls, bins = np.arange(0, 120, 3))

    # info: insert every sequence in the net for its size.
    #     Thus all sequences with one size are in one net
    #     Note that dividing by 3 is used to convert 
    #     nucleic acids to amino acids.

    nets = {}
    for seqEl in seq:
        #seqElIndex = int((len(seqEl) - lsHist[1][0])/3)
        seqElIndex = int((len(seqEl))/3)

        try:    
            if len(nets[seqElIndex]) < lMax:
                nets[seqElIndex].append(seqEl)
        
        except:
            nets[seqElIndex] = []
            nets[seqElIndex].append(seqEl)
    nets = collections.OrderedDict(sorted(nets.items())) # does he sort the according to items or keys?
    return nets

def loadSequence(step, plotData, i = -1, isExtractNum = True, animal='human', src=None):
    """
    info: - load the nucleic acid sequences, generated from SONIA 
        (and stored in a file 'pre.txt' or 'post.txt')
          - cut of all acids after a certain length, defined by 
              the threshold "extractNum"
          - store the sequences in the list "seq" and the lengths of the 
        the sequences in "ls"
    input: step: 0 -> data before the Thymus selection (load from pre.txt)
                 1 -> data after the Thymus selection (load from post.txt)
        plotData: PlotData object, that contains the important data and parameters;
            for more detail see documentation in plotData.py
        i: int number of letters to be extracted from each string, e.g. if i = 18,
            then only the first 18 letters from each string are extracted, the rest
            is ignored
        isExtractedNum: Bool, that says, if only a limited number of 
            letters should be extracted; Thus, if isExtractNum == False, it makes no
            sense, to set i > -1, because everything is extracted anyway.
            If isExtractNum == True, i must be set to a certain value > -1.
        animal: 'human' -> loads human data
            'mouse' -> loads mouse data
        src: None: loads from the normal standard files
            not None: loads the sequences from this filepath
    output: a: list, containing all the loaded data
        a4: list, containing the nucleic acid sequences
        seq: list, containing the nucleic acid sequences first acids until 
            the upper threshold "extractNum"
        filename: string (either 'pre.txt' or 'post.txt')
        ls: list of length values for each sequence in seq
    """
    # info: define filename
    if step == 0:
        if animal == 'human':
            filename = '/media/paul/extern/pre.txt'
        if animal == 'mouse':
            filename = '/home/paul/Dokumente/save_copy/imnet-master/pre_mouse.txt'
    if step == 1:
        if animal == 'human':
            filename = '/home/paul/Dokumente/save_copy/imnet-master/post.txt'
        if animal == 'mouse':
            filename = '/home/paul/Dokumente/save_copy/imnet-master/post_mouse.txt'

    filename = src if not src == None else filename
    # info: check, if plotData has an Ns-array, then choose N = plotData.Ns[i]; alternatively choose N = plotData.N[i]
    if i > -1:
        try: 
            N = plotData.Ns[i]
        except: 
            print("\n Could not load element plotData.Ns[i]. Perhaps it does not exist?")
            N = -1
    else:
        N = plotData.N

    # info: make variables and lists as empty or zero objects
    index = 0
    a = list()
    a4 = list()
    seq = list()
    ls = list() # right place to define ls?

    # info: opening the list of generated sequences 
    # info: (sequences generated done with SONIA, 
    # info: which simulates VDJ recombination)
    
    with open(filename) as f:
        f_lines = f.readlines()
        if len(f_lines) < N:
            print("\n Warning: There seem to be less than the requested \
                    N lines in the file. This can lead to errors, e.g. \
                    unknown CUDA error.")
        for line in f_lines:
            if index < N:
                a.append(line.split('\t'))
                # info: attach the fourth element  to a4
                #     the txt data consists of 4 columns. The 4th column
                #     contains the nucleic acid strings, which are relevant
                #     for our analysis. In order to load
                #     them, we can extract the fourth
                #     element of each line.
                #     Alternatively we can extract the first element 
                #     (1th column), which contains the amino acid 
                #     representation
                a4.append(a[index][0].replace('\n', ''))
                if isExtractNum == True:
                    seq.append(a4[index][:plotData.extractNum])
                else:
                    seq.append(a4[index][:])
                ls.append(len(seq[index]))
                
                index += 1
    return a, a4, seq, filename, ls

def get_stuff(filename, count, start_index):
    """
    Info: 
    Args: 
    Returns:
    """
    # TODO: This function is from the internet, should we 
    #     see that as intellectual property of someone?
    with open(filename, 'r') as infile:
             reader = csv.reader(infile, delimiter='\t')
             num = 0
             for idx, row in enumerate(reader):
                 if idx >= start_index-1:
                     if num >= count:
                         return
                 else:
                     yield row
                     num += 1

def load_easy(path = None, l = 100000):
    """
    Info: load fast or so that you can scale
    Args: path: path, where the file is
        l: number of lines to be loaded (if None, all is loaded)
    Returns: seq: list of CDR3 sequences as strings
    """
    print("\n begin load_easy ...")
    if path == None:
        path = "sequences/sequence_1.txt"
    print("\n info: loading from ", path) 
    with open(path, 'r') as f:
        #l = 10**3#10**6
        #N = 4*l
        #lines = f.read().split()
        #lines = f.readlines()
        print("\n AA")
        if l == None:
            lines = f.read().split()
        else:
            lines = get_stuff(path, 0, l+1)
            lines = list(lines)
        N = 4*len(lines)
        #print("\n -- lines[:min(100, l)]: ", lines[:min(100, len(lines))])
        #print("\n len(lines): ", len(lines), "; N: ", N)
        if len(lines) < N:
            print("\n Warning: len(lines) < N; N: ", N, "; len(lines): ", len(lines))
        #print("\n B, lines[:100]: ", lines[:100])
        ar = np.array(lines[:N])
        l_new = int(N/4)
        ar_new = np.reshape(ar, [l_new, 4])
        seq = ar_new[:, 0]
    print("\n load_easy finished, now returning seq ...")
    return seq

def make_sub_graphs(G):
    """
    info: make list of connected subgraphs of G (subgraph with the largest number of nodes) 
    input: G: graph
    output: sub_graphs: list of subgraphs, sorted according to size
    """
    # info: make sorted list of subgraphs
    sub_graphs = sorted(list(G.subgraph(c) for c in nx.connected_components(G)), key = len, reverse = True)

    # info: if sub_graphs is empty , make a base graph
    if len(sub_graphs) == 0:
        sub_graphs = list()
        sub_graphs.append(nx.Graph())
        sub_graphs[0].add_node('FillNode')

    return sub_graphs

def largest_sub_graph(G):
    """
    info: choose the largest component
    input: G: a graph
    output: Gc: largest component of G as subgraph
    """
    Gcc = sorted(nx.connected_components(G), key=len, reverse=True)
    
    Gc = G.subgraph(Gcc[0])
    
    return Gc
    
def append_zero(innerData):
    """
    info: append 0 to all members of an innerData - object
    input: innerData: an object, containing all the important network properties
    output: innerData: new object, containing the updated members
    """

    innerData.lengths.append(0)
    innerData.clustersizes.append(0)
    innerData.diameters.append(0)
    innerData.centers.append(0)
    innerData.eccentrities.append(0)
    innerData.degreeMean.append(0)
    return innerData

def append_values(innerData, GSub, G):
    """
    info: calculate the network properties of a graph "GSub" and append them to the corresponding lists in the innerData object
    input: innerData: object, that stores  network properties of many graphs in lists
    output: innerData: new object with updated lists
    """
    
    innerData.lengths.append(len(list(G.nodes())))
    innerData.clustersizes.append(len(GSub.nodes())/len(G.nodes()))
    # info: diameters ignored because of computation time
    #innerData.diameters.append(nx.diameter(GSub))
    innerData.diameters.append(-1)
    # info: centers ignored because of computation time
    #centerList = nx.center(GSub)
    #innerData.centers.append(centerList[0])
    innerData.centers.append(-1)
    innerData.eccentrities.append(0)
     # info: calculate the mean degree
    innerData.degreeMean.append(degree_mean(GSub))
    return innerData

def mean_and_error_of(plotData):
    """
    info: calculate the average values of the network properties, using the data in "plotData"
    input: plotData: PlotData object, that contains the important data and parameters;
            for more detail see documentation in plotData.py
    output: plotData with updated members
    """
    # info: calculate the Means and the Errors

    lengthsAllMean = 0
    clustersizesAllMean = 0
    diametersAllMean = 0
    centersAllMean = 0
    eccentritiesAllMean = 0
    degreeMeanAllMean = 0

    lengthsAllDifSquare = 0
    clustersizesAllDifSquare = 0
    diametersAllDifSquare = 0
    centersAllDifSquare = 0
    eccentritiesAllDifSquare = 0
    degreeMeanAllDifSquare = 0

    lAll = len(plotData.clustersizesAllArray)
    #print("\n lAll: ", lAll)
    
    for i in range(len(plotData.clustersizesAllArray)):
        lengthsAllMean += np.divide(plotData.lengthsAllArray[i],lAll)
        clustersizesAllMean += np.divide(plotData.clustersizesAllArray[i],lAll)
        diametersAllMean += np.divide(plotData.diametersAllArray[i],lAll)
        #centersAllMean += np.divide(centersAllArray[i],lAll)
        #eccentritiesAllMean += np.divide(eccentritiesAllSumArray[i],lAll)
        degreeMeanAllMean += np.divide(plotData.degreeMeanAllArray[i],lAll)
        """
        print("\n plotData.clustersizesAllArray[i]: ", plotData.clustersizesAllArray[i])
        print("\n plotData.diametersAllArray[i]: ", plotData.diametersAllArray[i])
        print("\n plotData.degreeMeanAllArray[i]: ", plotData.degreeMeanAllArray[i])
        print("\n plotData.clustersizesAllArray[i]: ", plotData.clustersizesAllArray[i])
        """

    for i in range(len(plotData.clustersizesAllArray)):
        lengthsAllDifSquare += np.divide(plotData.lengthsAllArray[i]  \
            - lengthsAllMean, 2)
        clustersizesAllDifSquare += np.power(plotData.clustersizesAllArray[i] \
            - clustersizesAllMean, 2)
        diametersAllDifSquare += np.power(plotData.diametersAllArray[i] \
            - diametersAllMean, 2)
        #centersAllSumSquare += np.power(centersAllArray[i], 2)
        #eccentritiesAllSumSquare += np.power(eccentritiesAll[i], 2)
        degreeMeanAllDifSquare += np.power(plotData.degreeMeanAllArray[i] \
            - degreeMeanAllMean, 2)
    
    # info: store the mean values in the plotData object
    plotData.lengths = lengthsAllMean
    plotData.clustersizes = clustersizesAllMean
    plotData.diameters = diametersAllMean
    plotData.degreeMean = degreeMeanAllMean
    """
    print("\n clustersizes: ", plotData.clustersizes)
    print("\n diameters: ", plotData.diameters)
    print("\n degreeMean: ", plotData.degreeMean)
    """
    
    # info: store the error values in the plotData object
    plotData.lenghtsAllErr = np.sqrt(np.divide( \
        lengthsAllDifSquare,(lAll - 1)))
    plotData.clustersizesAllErr = np.sqrt(np.divide( \
        clustersizesAllDifSquare,(lAll - 1)))
    plotData.diametersAllErr = np.sqrt(np.divide( \
        diametersAllDifSquare,(lAll - 1)))
    #centersAllSumSquare = np.sqrt(np.divide( \
    #    centersAllSumSquare,(lAll - 1)))
    #eccentritiesAllSumSquare = np.sqrt(np.divide( \
    #    eccentritiesAllSumSquare,(lAll - 1)))
    plotData.degreeMeanAllErr = np.sqrt(np.divide( \
        degreeMeanAllDifSquare,(lAll - 1)))
    # Shouldn't we also set the Mean properties?

    return plotData

def make_all_data(plotData, innerData, sample): # check description
    """
    info: calculate the sum of the clusterproperties of innerData and 
        plotData (thus the mean over all samples can be calculated later) # right? # Where is the normalization?
    input: plotData: object, storing the network properties
        innerData: object, storing the network properties for one sample # right?
        sample: the current sample # right? Or better "samples" as input? Where is it used?
    output: plotData: 
    """

    #plotData.xArgsAllArray.append(innerData.xArgs)
    plotData.lengthsAllArray.append(innerData.lengths)
    plotData.clustersizesAllArray.append(innerData.clustersizes)
    plotData.diametersAllArray.append(innerData.diameters)
    plotData.centersAllArray.append(innerData.centers)
    plotData.eccentritiesAllArray.append(innerData.eccentrities)
    plotData.degreeMeanAllArray.append(innerData.degreeMean)
    
    return plotData

def convert_to_amino_acids(plotData, innerData): # check description
    """
    info: convert lengths to the aminoacids
    input: plotData: PlotData object, that contains the important data and parameters;
            for more detail see documentation in plotData.py
        innerData: objects, that store network properties in list # in more detail perhaps
    output: plotData: object, that stores the network properties, now with updated members
    """
    """
    xArgsNew = list()
    lengthsNew = list()
    clustersizesNew = list()
    diametersNew = list()
    eccentritiesNew = list()
    degreeMeanNew = list()

    print("\n len(innerData.xArgs): ", len(innerData.xArgs))
    
    for i in range(len(innerData.xArgs)):
        #if i%3 == 0:
        print("\n i: ", i)
        if True:
            xArgsNew.append(innerData.xArgs[i])
            lengthsNew.append(innerData.lengths[i])
            clustersizesNew.append(innerData.clustersizes[i])
            diametersNew.append(innerData.diameters[i])
            eccentritiesNew.append(innerData.eccentrities[i])
            degreeMeanNew.append(innerData.degreeMean[i])
    print("\n clustersizesNew: ", clustersizesNew)    
    """
    plotData.lengthsAllArray.append(innerData.lengths)
    plotData.clustersizesAllArray.append(innerData.clustersizes)
    plotData.diametersAllArray.append(innerData.diameters)
    #plotData.centersAllArray.append(centers)
    plotData.eccentritiesAllArray.append(innerData.eccentrities)
    plotData.degreeMeanAllArray.append(innerData.degreeMean)
    plotData.xArgsAllArray.append(innerData.xArgs)
   
    return plotData # that belongs here, right?

def degree_mean(GSub):
    """
    info: calculate the mean degree in the graph GSub
    input: GSub: graph
    output: degree_mean: mean degree number (real number)
    """
   
    degrees = np.fromiter(dict(GSub.degree()).values(), dtype=float)
    degree_mean = np.mean(degrees)
    return degree_mean

def makeDegreeDistribution(G): # I think the name does not describe, what that actually does; better change name
    """
    info: calculate a list, which contains the sizes (number of nodes) of each connected subgraph (thus each cluster)
    input: G: graph
    output: clustersizes: list with the lengths of all clusters in G
        c: list of degree values of each node
    """

    # info: make a list of degrees of all nodes
    a = G.degree()
    
    # info: b: list, where each element is an array: element 0 is the string and element 1 is the corresponding degree value
    b = list(a)

    # info: make list c, which only contains the degree values
    c = list()
    
    c = [el[1] for el in b]
    #for i in range(len(b)):
    #    # what is this element? ...; Is that later used?
    #    c.append(b[i][1])
    #    
    
    # info: make the clusters, clustersizes    
    clusters = list(nx.connected_components(G))
    #clustersizes = list()
    #for i in range(len(clusters)):
    #     clustersizes.append(len(clusters[i]))
    clustersizes = [len(list(cluster)) for cluster in clusters]
    return clustersizes, c

def makeNet(names, maxDist, threshold, gpu_l):
    """
    info: make graph, where each node denotes one element of names. 
        names is an array of strings.
        if the jaccard distance between the strings, corresponding 
        to two nodes, is lower than "maxDist", an edge is drawn
    input: names:[String]:  list of strings
        maxDist (real number from 0 to 1): maximum jaccard distance between two strings,
            which is required to form a bond
    output: g: created graph, where each string is a node and two nodes are
        connected via an edge, if there jaccard distance is smaller
        than maxDist
    """
    g = nx.Graph()

    #print("\n len(names) in makeNet: ", len(names))
    for i in range(len(names)):
        g.add_node(names[i])
    #g = values.map(netLoop) what's that?
    distance_matrix = jaccardUtil.jaccard(names, gpu_l, threshold)
    try:
        sz = distance_matrix.shape
        edgelist = list()
        for i in range(sz[0]):
            for j in range(i, sz[1]):
                #if distance_matrix[0][1] < threshold:
                if distance_matrix[i][j] < threshold:
                    edgelist.append([i, j])

        g.add_edges_from(edgelist)
    except:
        print("\n paul: distance_matrix is probably None; continuing anyway")
    #g = netLoop(g, names, maxDist)
    return g

def open_or_create(filename):
    """
    info: open a file with the name or path "filename",
        if the file does not yet exist, create a new 
        file and open it
    input: filename: name or path of the file,
        e.g. "example_data.txt" or "./example_folder/example_data.txt"
    output: datafile
    """
    try:
        datafile = open(filename, "w")
    except:
        datafile = open(filename, "x")
    return datafile



from matplotlib.colors import LinearSegmentedColormap
import numpy as np

#import matplotlib.pyplot as plt
from matplotlib import rcParams
import matplotlib.pyplot as plt

def plot(xs, ys, log=True, name=None):
    """
    Info: plot the array ys over the array xs with 
        some standard settings and save it
    Args: xs: array of x-values
        ys: array of y-values
        log: bool: True: log-log-plot; False: no log-plot
        name: name of the file to save to
    Returns: -
    """

    fig_size, params = design(factor=1)
    rcParams.update({'figure.figsize': fig_size})

    fig_1, ax_1 = plt.subplots()

    ax_1.plot(xs, ys, '.')
    if log == True:
        ax_1.set_xscale('log')
        ax_1.set_yscale('log')
    if name!=None:
        fig_1.tight_layout()
        fig_1.savefig(name, dpi=300, bbox_inches="tight")
    plt.show()

def background(ax_00):
    """
    info: make a gray color gradient for the background
    input: ax_00
    output: ax_00
    """

    cmap = LinearSegmentedColormap.from_list('mycmap', ['gray', 'white'])
    
    lx = 101
    ly = 101
    z = np.zeros([lx, ly])
    for i in range(lx):
        for j in range(ly):
            z[i][j] = i + j

    x = list(range(lx))
    y = list(range(ly))
    x = [el/100 for el in x]
    y = [el/100 for el in y]
    x, y = np.meshgrid(x, y)

    ax_00.pcolor(x, y, z, cmap=cmap, transform=ax_00.transAxes)

    return ax_00

def design(factor=2, factor_ratio=1):
    """
    info: 
    input:
    output: 
    """
    
    # set nice figure sizes
    fig_width_pt = 245*factor    # Get this from LaTeX using \showthe\columnwidth
    golden_mean = (np.sqrt(5.) - 1.) / 2.  # Aesthetic ratio
    ratio = golden_mean
    inches_per_pt = 1. / 72.27  # Convert pt to inches
    fig_width = fig_width_pt * inches_per_pt  # width in inches
    fig_height = fig_width*ratio*factor_ratio # height in inches
    fig_size = [fig_width, fig_height]
    rcParams.update({'figure.figsize': fig_size})
    
    params = { #'backend': 'ps',
            'font.family': 'serif',
            #'font.serif': 'Latin Modern Roman',
            #'font.family': 'sans serif',
            'font.serif': 'Helvetica',
            'font.size': 10,
            'axes.labelsize': 'medium',
            'axes.titlesize': 'medium',
            'legend.fontsize': 'medium',
            'xtick.labelsize': 'small',
            'ytick.labelsize': 'small',
            'savefig.dpi': 15,
            'text.usetex': True
    }
    
    # tell matplotlib about your params
    rcParams.update(params)
    
    # sphinx_gallery_thumbnail_number = 7
    
    #plt.rcParams['savefig.facecolor'] = "0.8"


    """
    params = { #'backend': 'ps',
        'font.family': 'serif',
        #'font.serif': 'Latin Modern Roman',
        #'font.family': 'sans serif',
        'font.serif': 'Helvetica',
        'font.size': 10,
        'axes.labelsize': 'medium',
        'axes.titlesize': 'medium',
        'legend.fontsize': 'medium',
        'xtick.labelsize': 'small',
        'ytick.labelsize': 'small',
        'savefig.dpi': 15,
        'text.usetex': True
    }
    ""
    rcParams['axes.labelsize'] = 50
    rcParams['xtick.labelsize'] = 46
    rcParams['ytick.labelsize'] = 46
    rcParams['legend.fontsize'] = 46
    rcParams['font.family'] = 'sans serif'
    rcParams['font.serif'] = ['Helvetica']
    rcParams['text.usetex'] = True
    rcParams['figure.figsize'] = 12, 8
    ""

    # set nice figure sizes
    fig_width_pt = 245    # Get this from LaTeX using \showthe\columnwidth
    golden_mean = (np.sqrt(5.) - 1.) / 2.  # Aesthetic ratio
    ratio = golden_mean
    inches_per_pt = 1. / 72.27  # Convert pt to inches
    fig_width = fig_width_pt * inches_per_pt  # width in inches
    fig_height = fig_width*ratio  # height in inches
    fig_size = [fig_width, fig_height]
    """

    return fig_size, params 

if __name__ == "__main__": 
    print("\n doing main")
    design()



from numba import jit, njit, vectorize
import time
import matplotlib.pyplot as plt
import numpy as np

import sys
import argparse 

from concurrent.futures import ThreadPoolExecutor

from numba import cuda

from . import funcDictionary as dic
#import funcDictionaryLevenshtein as dicLeven
from . import plotData

import psutil
from scipy import sparse
import pandas as pd
import math

from .graph_support import *

def gpu_sparse(strings1, strings2, num_strings, blockspergrid, threadsperblock):
    """
    Info: calculate the sparse Levenshtein distance matrix of 
        two arrays "strings1" and "strings2"  on the GPU
    Args: strings1: first array of strings, encoded in ASCII numbers, 
            length_of_array = num_strings * length_per_string, with 
            length_per_string = 30
        strings2: first array of strings, encoded in ASCII numbers
        num_strings: number of strings in each array, usually 4*10**3, 
        blockspergrid: a parameter for the GPU calculation
        threadsperblock: a parameter for the GPU calculation
    Returns: h_sparse: list of indices, where the matrix entry, is 1. 
            The entry is 1, if and only if the Levenshtein distance between 
            the two corresponding strings is smaller than or equal to the 
            threshold distance max_ld = 2.
    """
    # TODO: check, if our condition is <2 or <=2
    strings1 = strings1.astype(dtype=np.int8)
    strings2 = strings2.astype(dtype=np.int8)

    dist = np.zeros(num_strings**2)#**2
    dist = dist.astype(dtype=np.int8)

    d_strings1 = cuda.to_device(strings1)
    d_strings2 = cuda.to_device(strings2)

    d_num_strings = cuda.to_device(num_strings)
    d_dist = cuda.to_device(dist)

    max_ld = 2
    d_max_ld = cuda.to_device(max_ld)

    cuda.synchronize()
    levenshtein[blockspergrid, threadsperblock](d_strings1, d_strings2, num_strings, d_dist, max_ld)
    cuda.synchronize()
    h_dist = d_dist.copy_to_host()
    h_sparse = np.asarray(np.where(h_dist!=0), dtype=np.int32)
    h_sparse = h_sparse[0]
    return h_sparse

def total_idx(el, N_part, i, j, len_x, len_y):
    """
    Info: calculate the total one-number-index from the one-number-index
        in the small block matrix with sidelength N_part
    Args: el: one-number-index in the long block
        N_part: sidelength of single block (usually 4*10**3) 
        i: row index of the block
        j: column index of the block
        len_x: number of blocks in one row
        len_y: number of blocks in one column
    Returns: idx: Total index, one number
    """
    offset_index = len_x * i * N_part**2
    el_y = int(el/N_part)
    el_x = int(el%N_part)
    el_A = el_x * N_part * len_y
    el_B = el_y + j*N_part
    idx = offset_index + el_A + el_B
    return idx

def adjacency_matrix(seq, name="sparse.txt", idx_max=0, name_params="data/sparse_params.txt", N_part=None, len_xy=None, direct_output=False):
    """
    Info: calculating the entire adjacency matrix and save it
    Args: seq: the array of strings
        name: name of the txt file to save the one-number-index
        idx_max: number of sequences, which are replaced by virus 
            recognizing sequences (0, except if you do virus analysis)
        name_params: name of the txt file to save idx_max.
        N_part: sidelength of a single block, default: 4*10**3
        len_xy: number of single blocks in x and y direction respectively, default: 10
        direct_output: if True, return the data directly (only suitable for small data), default: False
    Returns: direct_output: returns the adjacency sparse matrix, if direct_output == True, else returns empty array
    """
    with_log = False

    if not idx_max == 0:
        with open(name_params, 'w') as f:
            np.savetxt(f, np.array([idx_max]), fmt='%i')

    Ns = np.multiply([0.1, 0.5, 1.0, 2.0, 3.0, 4.0, 5.0, 10.0], 10**3)
    maxVals = [1, 2, 3, 4, 5]
    #N_part = 4*10**3
    num_strings = N_part
    threadsperblock = 256#256# 24
    blockspergrid = ((num_strings)**2 + (threadsperblock - 1)) // threadsperblock

    if with_log:
        print("\n starting ascii ...")
    strings_1, strings_2 = convert_to_ascii(seq, seq, 30)
    if with_log:
        print("\n finished ascii")

    #@cuda.jit('void(int8[:], int8, int8[:])')
    tA = time.time()
    
    all_array = np.array([])

    # TODO: make sure, you always iterate over the entire length of seq
    with open(name, 'w+') as f:
        len_x, len_y = len_xy, len_xy
        for i in range(len_x):#250
            if with_log: 
                print("\n now setting to zero ...")
            h_total = np.array([])
            if with_log: 
                print("\n ... finished setting to zero")
            for j in range(i, len_y):#250
                if with_log: 
                    print("\n i: ", i, "j: ", j)
                strings1 = strings_1[(i*N_part*30):((i+1)*N_part*30)]
                strings2 = strings_2[(j*N_part*30):((j+1)*N_part*30)]
                h_sparse = gpu_sparse(strings1, strings2, num_strings, blockspergrid, threadsperblock)
                if with_log: 
                    print("\n np.shape(h_sparse): ", np.shape(h_sparse), "; np.shape(h_total): ", np.shape(h_total))

                # info: addding the offset index of the block
                #offset_index = (len_y * i + j) * N_part**2
                #h_sparse = [el + offset_index for el in h_sparse]
                h_sparse = [total_idx(el, N_part, i, j, len_x, len_y) for el in h_sparse]
                # info: removing self edges
                h_sparse = contract_array(h_sparse, idx_max, N_part*len_y)
                h_total = np.concatenate([h_total, h_sparse])
            all_array = np.concatenate([all_array, h_total]) if direct_output else all_array
            s = "\n " + str(h_total)
            if with_log: 
                print("\n h_total[0]: ", h_total[0], "; h_total[1]: ", \
                        h_total[1], "; el%N_part: ", h_total[0]%N_part, \
                        "; int(el/N_part): ", int(h_total[1]/N_part))
            #-h_total = [[el%(N_part*len_x), int(el/(N_part*len_x))] for el in h_total]
            #f.write(s)
            np.savetxt(f, h_total, fmt='%i') # perh TODO: check that this really saves the entire matrix correctly
    tB = time.time()
    if with_log: 
        print("\n dt: ", tB - tA) 
        print("\n blockspergrid: ", blockspergrid)
    
    return all_array

if __name__ == "__main__":
    fig1, ax1 = plt.subplots()
    
    with_log = False

    #Ns = [1*10**1, 2*10**1]#, 10**4, 10**5]
    Ns = [10**1]
    t_py = []
    t_gpu = []
    
    # TODO: establish command line argument taking
    parser = argparse.ArgumentParser()
    parser.add_argument('--N', help="number of sequences to be loaded, default 10**6")
    parser.add_argument('--N_part', help="number of sequences per single block, \
            default 4*10**3, better don't change that")
    parser.add_argument('--len_xy', help="number of single blocks in x and y direction respectively. \
            number of processed (not loaded) sequences: len_xy * N_part")
    parser.add_argument('--name', help="name of the single index edge list file")
    parser.add_argument('--name_params', help="name of the name_params file")
    parser.add_argument('--with_pathogens', help="1: with pathogens, 0 (default): without pathogens")
    parser.add_argument('--src', help="a special source name from where to load sequences")
    # TODO: make sure that the path is also set correctly when 
    #     with_pathogens=1, currently he just saves under the 
    #     default name and path
    args = parser.parse_args()

    parser_N = int(args.N) if not args.N == None else 10**6
    parser_N_part = int(args.N_part) if not args.N_part == None else 4*10**3
    parser_len_xy = int(args.len_xy) if not args.len_xy == None else 10
    parser_name = args.name if not args.name == None else "data/sparse.txt"
    parser_name_params = args.name_params if not args.name_params == None else "data/sparse_params.txt"
    with_pathogens = True if args.name_params == 1 else 0
    src_name = args.src

    for N in Ns:
        plotData.clusterMinLen = 1
        plotData.N = parser_N#10**6#4*10**3
        plotData.min_ldVal = -1
        plotData.maxVal = 3
        plotData.max_ldVal = plotData.maxVal
        gpu_l = 8000 #5000
        step = 0
        a, a4, seq, filename, ls = dic.loadSequence(step, plotData, isExtractNum=False, src=src_name)

        #with_pathogens = True
        if with_pathogens:
            pathogens_savenames = ["Influenza", "HIV", "Alzheimer", \
                    "Parkinson", "Tuberculosis", "COVID-19"]
            pathogens = ["Influenza", "Human immunodeficiency virus (HIV)", \
                    "Alzheimer's disease", "Parkinson disease", \
                    "M.Tuberculosis", "COVID-19"]
            for pathogen, save_name in zip(pathogens, pathogens_savenames):
                index = 0
                if with_log: 
                    print("\n ------------------- pathogen index: ", pathogens.index(pathogen))
                seq, idx_max = insert_pathogens(seq, pathogen)# if with_pathogens == True else seq   

                name = "data/" + save_name + "_small.txt"
                name_params = "data/" + save_name + "_params.txt"
                #-print("\n seq[:500]: ", seq[:500])
                _ = adjacency_matrix(seq, name=name, idx_max=idx_max, name_params=name_params, N_part=parser_N_part, len_xy=parser_len_xy)
        else:
            name = parser_name#"data/sparse.txt"
            _ = adjacency_matrix(seq, name=name, N_part=parser_N_part, len_xy=parser_len_xy)
            if with_log: 
                print("\n saved under name: ", name)



from setuptools import setup

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="TCRnumba",
    version='0.1.8',
    description="some description",
    url="https://github.com/Paul44444/capybaras_test",
    author='Paul',
    author_email='richtepp@physik.hu-berlin.de',
    license='BSD 2-clause',
    packages=['TCRnumba'],
    #install_requires=["mpi4py>=2.0","numpy", "numba", "scipy"], #TODO: finish this list
    install_requires=["mpi4py>=2.0","numpy", "numba", "scipy"], 
    #classifiers=[],
    
    long_description=long_description,
    long_description_content_type='text/markdown',
    
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
) 



# info: perform analysis of the private and the public parts
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

from matplotlib import rcParams
from . import graphics

import sys
sys.path.append("./public_analysis/")
import plot1 as p1
import plot2 as p2
import plot3 as p3
import plot4 as p4

from . import funcDictionary as dic
#import funcDictionaryLevenshtein as dicLeven
from . import plotData as pD

import networkx as nx
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import rcParams
import os

import time

import networkx as nx
import sys

#import imnet
import random

import matplotlib.patches as mpatches

from matplotlib.patches import Polygon
from matplotlib.collections import PatchCollection
from matplotlib.colors import ListedColormap
from collections import OrderedDict

import seaborn as sns

# info: "graphics" is a script, written by myself
from . import graphics
import matplotlib.gridspec as gridspec

plotData = pD.PlotDataLevenshtein()
plotData.clusterMinLen = 1

# info: plotData.N: the total number of sequences, which is loaded
plotData.N = 10**5

# info: size of one sample (and then they compare many samples
#     in order to perform the private vs. public analysis)
sample_size = 8*10**3 # 40*10**3 seems to be possible
samples = 5

plotData.min_ldVal = -1
plotData.maxVal = 3
plotData.max_ldVal = plotData.maxVal
step = 0

maxVals = [1, 2, 3, 4, 5]

#------------------------------------------------------------------

fig_size, params = graphics.design()
rcParams.update({'figure.figsize': fig_size})

# tell matplotlib about your params
rcParams.update(params)

#fig, ax = plt.subplots(nrows = 2, ncols = 2)
fig = plt.figure()
fig_dummy = plt.figure()

gs00 = gridspec.GridSpec(1, 1)
gs01 = gridspec.GridSpec(1, 1)
gs11 = gridspec.GridSpec(1, 1)
gs10A = gridspec.GridSpec(1, 1)
gs10B = gridspec.GridSpec(1, 1)

ax00 = fig.add_subplot(gs00[0])
ax01 = fig.add_subplot(gs01[0])
ax10 = fig.add_subplot(gs10A[0])
#-ax_public_simple = fig.add_subplot(gs10B[0])
ax_public_simple = fig_dummy.add_subplot(gs10B[0])
ax11 = fig.add_subplot(gs11[0])
ax = [[ax00, ax01], [ax10, ax11]]

gs00.tight_layout(fig, rect=[0.5, 0.5, 1.0, 1.0])
gs01.tight_layout(fig, rect=[0.0, 0.0, 0.5, 0.5])

gs11.tight_layout(fig, rect=[0.5, 0.0, 1.0, 0.5])
gs10A.tight_layout(fig, rect=[0.0, 0.5, 0.5, 1.0])

plt.tight_layout()

# set nice figure sizes
fig_width_pt = 245/1.00    # Get this from LaTeX using \showthe\columnwidth
golden_mean = (np.sqrt(5.) - 1.) / 2.  # Aesthetic ratio
ratio = golden_mean
inches_per_pt = 1. / 72.27  # Convert pt to inches
fig_width = fig_width_pt * inches_per_pt  # width in inches
fig_height = fig_width*ratio  # height in inches
fig_size = [fig_width, fig_height]
rcParams.update({'figure.figsize': fig_size})

# tell matplotlib about your params
rcParams.update(params)

fig_bars, ax_a = plt.subplots()
fig_bars = plt.figure()
ax_a = plt.subplot2grid((2, 2), (0, 0), rowspan=2, colspan=2)

for name in ["not_important"]: 
    #-ax_a, ax_b, ax_c, _, ax[0][0], collection = p1.plot1(ax_a, ax_b, ax_c, ax_d, ax[0][0])
    ax_a, ax[0][0], collections = p1.plot1(ax_a, ax[0][0])
    #_, ax[0][0], collection = p1b.plot1(ax_a, ax[0][0])
    #ax[0][1] = p2.plot2(ax[0][1], collections)
    sample_size = 10**5#8*10**3
    ax[1][1], g_vdj, seqss = p4.plot4(ax[1][1], sample_size=sample_size)
    print("\n --- len(seqss): ", len(seqss['mouse']), "; len(seqss): ", len(seqss['mouse']))
    print(str(os.path.realpath(__file__)), ": done plot4, starting plot3 ...")
    #print("\n /home/paul/Documents/imnet-master2/public_analysis.py: done plot4, starting plot3 ...") 
    ax[1][0], ax_a, ax_public_simple, ax_b = p3.plot3(ax[1][0], g_vdj, seqss, ax_a, ax_public_simple, samples_big=2)
    print("\n /home/paul/Documents/imnet-master2/public_analysis.py: done this part")
    # info: ax[1][0] - somehow cliquesize of the network created from certain frequency networks 

fig.tight_layout()
fig_bars.tight_layout()
fig.savefig("public_analysis.png", dpi=300, bbox_inches="tight")
fig_bars.savefig("parameters.png", dpi=300, bbox_inches="tight")
#fig_public_simple.savefig("public_simple.png", dpi=300, bbox_inches="tight")
plt.show()
